@@ -43,16 +43,25 @@ from cvxpy.reductions.reduction import Reduction
 from cvxpy.reductions.solvers import defines as slv_def
 from cvxpy.reductions.solvers.constant_solver import ConstantSolver
 from cvxpy.reductions.solvers.solver import Solver
-from cvxpy.settings import ECOS, PARAM_THRESHOLD
+from cvxpy.settings import CLARABEL, ECOS, PARAM_THRESHOLD
 from cvxpy.utilities.debug_tools import build_non_disciplined_error_msg
 
 DPP_ERROR_MSG = (
     "You are solving a parameterized problem that is not DPP. "
     "Because the problem is not DPP, subsequent solves will not be "
     "faster than the first one. For more information, see the "
-    "documentation on Discplined Parametrized Programming, at\n"
-    "\thttps://www.cvxpy.org/tutorial/advanced/index.html#"
-    "disciplined-parametrized-programming"
+    "documentation on Disciplined Parametrized Programming, at "
+    "https://www.cvxpy.org/tutorial/dpp/index.html"
+)
+
+ECOS_DEP_DEPRECATION_MSG = (
+    """
+    You specified your problem should be solved by ECOS. Starting in
+    CXVPY 1.6.0, ECOS will no longer be installed by default with CVXPY.
+    Please either add an explicit dependency on ECOS or switch to our new
+    default solver, Clarabel, by either not specifying a solver argument
+    or specifying ``solver=cp.CLARABEL``.
+    """
 )
 
 ECOS_DEPRECATION_MSG = (
@@ -145,20 +154,20 @@ def _reductions_for_problem_class(problem, candidates, gp: bool = False, solver_
     if type(problem.objective) == Maximize:
         reductions += [FlipObjective()]
 
-    use_quad = True if solver_opts is None else solver_opts.get('use_quad_obj', True)
-    if _solve_as_qp(problem, candidates) and use_quad:
-        reductions += [CvxAttr2Constr(), qp2symbolic_qp.Qp2SymbolicQp()]
-    else:
-        # Canonicalize it to conic problem.
-        if not candidates['conic_solvers']:
-            raise SolverError("Problem could not be reduced to a QP, and no "
-                              "conic solvers exist among candidate solvers "
-                              "(%s)." % candidates)
-
+    # Special reduction for finite set constraint, 
+    # used by both QP and conic pathways.
     constr_types = {type(c) for c in problem.constraints}
     if FiniteSet in constr_types:
         reductions += [Valinvec2mixedint()]
 
+    use_quad = True if solver_opts is None else solver_opts.get('use_quad_obj', True)
+    valid_qp = _solve_as_qp(problem, candidates) and use_quad
+    valid_conic = len(candidates['conic_solvers']) > 0
+    if not valid_qp and not valid_conic:
+        raise SolverError("Problem could not be reduced to a QP, and no "
+                            "conic solvers exist among candidate solvers "
+                            "(%s)." % candidates)
+
     return reductions
 
 
@@ -245,8 +254,12 @@ def construct_solving_chain(problem, candidates,
         # Canonicalize as a QP
         solver = candidates['qp_solvers'][0]
         solver_instance = slv_def.SOLVER_MAP_QP[solver]
-        reductions += [QpMatrixStuffing(canon_backend=canon_backend),
-                       solver_instance]
+        reductions += [
+            CvxAttr2Constr(reduce_bounds=not solver_instance.BOUNDED_VARIABLES), 
+            qp2symbolic_qp.Qp2SymbolicQp(),
+            QpMatrixStuffing(canon_backend=canon_backend),
+            solver_instance,
+        ]
         return SolvingChain(reductions=reductions)
 
     # Canonicalize as a cone program
@@ -300,7 +313,13 @@ def construct_solving_chain(problem, candidates,
 
     # Here, we make use of the observation that canonicalization only
     # increases the number of constraints in our problem.
-    has_constr = len(cones) > 0 or len(problem.constraints) > 0
+    var_domains = sum([var.domain for var in problem.variables()], start = [])
+    has_constr = len(cones) > 0 or len(problem.constraints) > 0 or len(var_domains) > 0
+
+    if PSD in cones \
+            and slv_def.DISREGARD_CLARABEL_SDP_SUPPORT_FOR_DEFAULT_RESOLUTION \
+            and specified_solver is None:
+        candidates['conic_solvers'] = [s for s in candidates['conic_solvers'] if s != CLARABEL]
 
     for solver in candidates['conic_solvers']:
         solver_instance = slv_def.SOLVER_MAP_CONIC[solver]
@@ -312,6 +331,7 @@ def construct_solving_chain(problem, candidates,
         unsupported_constraints = [
             cone for cone in cones if cone not in supported_constraints
         ]
+
         if has_constr or not solver_instance.REQUIRES_CONSTR:
             if ex_cos:
                 reductions.append(Exotic2Common())
@@ -327,12 +347,12 @@ def construct_solving_chain(problem, candidates,
                 problem.objective.expr.has_quadratic_term()
             reductions += [
                 Dcp2Cone(quad_obj=quad_obj),
-                CvxAttr2Constr(),
+                CvxAttr2Constr(reduce_bounds=not solver_instance.BOUNDED_VARIABLES),
             ]
             if all(c in supported_constraints for c in cones):
-                # Raise a warning if ECOS is used without being specified
-                # by the user.
-                if solver == ECOS and specified_solver is None:
+                if solver == ECOS and specified_solver == ECOS:
+                    warnings.warn(ECOS_DEP_DEPRECATION_MSG, FutureWarning)
+                elif solver == ECOS and specified_solver is None:
                     warnings.warn(ECOS_DEPRECATION_MSG, FutureWarning)
                 # Return the reduction chain.
                 reductions += [
