@@ -26,6 +26,7 @@ from numpy import linalg as LA
 import cvxpy as cp
 import cvxpy.settings as s
 from cvxpy import Minimize, Problem
+from cvxpy.atoms.errormsg import SECOND_ARG_SHOULD_NOT_BE_EXPRESSION_ERROR_MESSAGE
 from cvxpy.expressions.constants import Constant, Parameter
 from cvxpy.expressions.variable import Variable
 from cvxpy.reductions.solvers.defines import INSTALLED_MI_SOLVERS
@@ -204,6 +205,13 @@ class TestAtoms(BaseTest):
         self.assertTrue(copy.args[0] is self.y)
         self.assertEqual(copy.get_data(), atom.get_data())
 
+        # Error message when geo_mean used incorrectly.
+        with pytest.raises(
+                TypeError, 
+                match=SECOND_ARG_SHOULD_NOT_BE_EXPRESSION_ERROR_MESSAGE
+            ):
+            cp.geo_mean(self.x, self.y)
+
     # Test the harmonic_mean class.
     def test_harmonic_mean(self) -> None:
         atom = cp.harmonic_mean(self.x)
@@ -229,6 +237,10 @@ class TestAtoms(BaseTest):
         self.assertEqual(atom.sign, s.NONNEG)
         expr = cp.norm(self.A, 2, axis=0)
         self.assertEqual(expr.shape, (2,))
+        expr = cp.norm(self.A, 2, axis=0, keepdims=True)
+        self.assertEqual(expr.shape, (1, 2))
+        expr = cp.norm(self.A, 2, axis=1, keepdims=True)
+        self.assertEqual(expr.shape, (2, 1))
 
         atom = cp.pnorm(self.x, p='inf')
         self.assertEqual(atom.shape, tuple())
@@ -388,6 +400,16 @@ class TestAtoms(BaseTest):
             cp.min(self.x, self.x)  # a common erroneous use-case
         self.assertEqual(str(cm.exception), cp.min.__EXPR_AXIS_ERROR__)
 
+        # Test canonicalization with keepdims=True
+        # https://github.com/cvxpy/cvxpy/pull/2419
+        X = cp.Variable((2, 3))
+        X_val = np.arange(6).reshape((2, 3))
+        c = np.ones((1, 3))
+        expr = cp.min(X, axis=0, keepdims=True)
+        obj = cp.Maximize(cp.sum(expr + c))
+        prob = cp.Problem(obj, [X == X_val])
+        prob.solve()
+
     # Test sign logic for maximum.
     def test_maximum_sign(self) -> None:
         # Two args.
@@ -588,6 +610,41 @@ class TestAtoms(BaseTest):
         self.assertItemsAlmostEqual(b_reshaped, X_reshaped.value)
         self.assertItemsAlmostEqual(b, X.value)
 
+    def test_reshape_negative_one(self) -> None:
+        """
+        Test the reshape class with -1 in the shape.
+        """
+
+        expr = cp.Variable((2, 3))
+        numpy_expr = np.ones((2, 3))
+        shapes = [(-1, 1), (1, -1), (-1, 2), -1, (-1,)]
+        expected_shapes = [(6, 1), (1, 6), (3, 2), (6,), (6,)]
+
+        for shape, expected_shape in zip(shapes, expected_shapes):
+            expr_reshaped = cp.reshape(expr, shape)
+            self.assertEqual(expr_reshaped.shape, expected_shape)
+
+            numpy_expr_reshaped = np.reshape(numpy_expr, shape)
+            self.assertEqual(numpy_expr_reshaped.shape, expected_shape)
+
+        with pytest.raises(ValueError, match="Cannot reshape expression"):
+            cp.reshape(expr, (8, -1))
+
+        with pytest.raises(AssertionError, match="Only one"):
+            cp.reshape(expr, (-1, -1))
+
+        with pytest.raises(ValueError, match="Invalid reshape dimensions"):
+            cp.reshape(expr, (-1, 0))
+
+        with pytest.raises(AssertionError, match="Specified dimension must be nonnegative"):
+            cp.reshape(expr, (-1, -2))
+
+        A = np.array([[1, 2, 3], [4, 5, 6]])
+        A_reshaped = cp.reshape(A, -1, order='C')
+        assert np.allclose(A_reshaped.value, A.reshape(-1, order='C'))
+        A_reshaped = cp.reshape(A, -1, order='F')
+        assert np.allclose(A_reshaped.value, A.reshape(-1, order='F'))
+
     def test_vec(self) -> None:
         """Test the vec atom.
         """
@@ -643,6 +700,22 @@ class TestAtoms(BaseTest):
         self.assertFalse(expr.is_psd())
         self.assertFalse(expr.is_nsd())
 
+    def test_diag_offset(self) -> None:
+        """Test matrix to vector on scalar matrices"""
+        test_matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
+        test_vector = np.array([1, 2, 3])
+        offsets = [0, 1, -1, 2]
+        for offset in offsets:
+            a_cp = cp.diag(test_matrix, k=offset)
+            a_np = np.diag(test_matrix, k=offset)
+            A_cp = cp.diag(test_vector, k=offset)
+            A_np = np.diag(test_vector, k=offset)
+            self.assertItemsAlmostEqual(a_cp.value, a_np)
+            self.assertItemsAlmostEqual(A_cp.value, A_np)
+
+        X = cp.diag(Variable(5), 1)
+        self.assertEqual(X.size, 36)
+
     def test_trace(self) -> None:
         """Test the trace atom.
         """
@@ -685,16 +758,15 @@ class TestAtoms(BaseTest):
                          "Argument to upper_tri must be a square matrix.")
 
     def test_vec_to_upper_tri(self) -> None:
-        from cvxpy.atoms.affine.upper_tri import vec_to_upper_tri
         x = Variable(shape=(3,))
-        X = vec_to_upper_tri(x)
+        X = cp.vec_to_upper_tri(x)
         x.value = np.array([1, 2, 3])
         actual = X.value
         expect = np.array([[1, 2], [0, 3]])
         assert np.allclose(actual, expect)
         y = Variable(shape=(1,))
         y.value = np.array([4])
-        Y = vec_to_upper_tri(y, strict=True)
+        Y = cp.vec_to_upper_tri(y, strict=True)
         actual = Y.value
         expect = np.array([[0, 4], [0, 0]])
         assert np.allclose(actual, expect)
@@ -703,9 +775,28 @@ class TestAtoms(BaseTest):
                              [0, 0, 0, 21],
                              [0, 0, 0, 0]])
         a = np.array([11, 12, 13, 16, 17, 21])
-        A_actual = vec_to_upper_tri(a, strict=True).value
+        A_actual = cp.vec_to_upper_tri(a, strict=True).value
         assert np.allclose(A_actual, A_expect)
 
+        with pytest.raises(ValueError, match="must be a triangular number"):
+            cp.vec_to_upper_tri(cp.Variable(shape=(4)))
+
+        with pytest.raises(ValueError, match="must be a triangular number"):
+            cp.vec_to_upper_tri(cp.Variable(shape=(4)), strict=True)
+
+        with pytest.raises(ValueError, match="must be a vector"):
+            cp.vec_to_upper_tri(cp.Variable(shape=(2, 2)))
+
+        # works with row vectors
+        assert np.allclose(
+            cp.vec_to_upper_tri(np.arange(6)).value, 
+            cp.vec_to_upper_tri(np.arange(6).reshape(1, 6)).value
+        )
+
+        # works with scalars
+        assert np.allclose(cp.vec_to_upper_tri(1, strict=True).value, np.array([[0, 1], [0, 0]]))
+
+
     def test_huber(self) -> None:
         # Valid.
         cp.huber(self.x, 1)
@@ -791,6 +882,10 @@ class TestAtoms(BaseTest):
         copy = atom.copy()
         self.assertTrue(type(copy) is type(atom))
 
+        # Check that sum_largest is PWL so can be canonicalized as a QP.
+        atom = cp.sum_largest(self.x, 2)
+        assert atom.is_pwl()
+
     def test_sum_smallest(self) -> None:
         """Test the sum_smallest atom and related atoms.
         """
@@ -804,6 +899,10 @@ class TestAtoms(BaseTest):
         self.assertEqual(str(cm.exception),
                          "Second argument must be a positive integer.")
 
+        # Check that sum_smallest is PWL so can be canonicalized as a QP.
+        atom = cp.sum_smallest(self.x, 2)
+        assert atom.is_pwl()
+
     def test_index(self) -> None:
         """Test the copy function for index.
         """
@@ -858,6 +957,21 @@ class TestAtoms(BaseTest):
         self.assertEqual(str(cm.exception),
                          "The arguments to conv must resolve to vectors.")
 
+        # Test with parameter input.
+        # https://github.com/cvxpy/cvxpy/issues/2218
+        x = cp.Variable()
+        p = cp.Parameter()
+        problem = cp.Problem(cp.Minimize(cp.conv(p, x)), [0 <= x, x <= 1])
+
+        p.value = -1.0
+        result = problem.solve()
+        self.assertAlmostEqual(result, -1)
+        self.assertAlmostEqual(x.value, 1)
+
+        problem = cp.Problem(cp.Minimize(cp.conv(p, x)), [0 <= x, x <= 1])
+        with pytest.raises(cp.DPPError):
+            problem.solve(enforce_dpp=True)
+
     def test_kron_expr(self) -> None:
         """Test the kron atom.
         """
@@ -892,6 +1006,90 @@ class TestAtoms(BaseTest):
         with pytest.raises(ValueError, match="scalar or 1D"):
             cp.convolve([[0, 1], [0, 1]], self.x)
 
+        # Test with parameter input.
+        # https://github.com/cvxpy/cvxpy/issues/2218
+        x = cp.Variable()
+        p = cp.Parameter()
+        problem = cp.Problem(cp.Minimize(cp.convolve(p, x)), [0 <= x, x <= 1])
+
+        p.value = -1.0
+        result = problem.solve(canon_backend=cp.CPP_CANON_BACKEND)
+        self.assertAlmostEqual(result, -1)
+        self.assertAlmostEqual(x.value, 1)
+
+        problem = cp.Problem(cp.Minimize(cp.convolve(p, x)), [0 <= x, x <= 1])
+        with pytest.raises(cp.DPPError):
+            problem.solve(enforce_dpp=True)
+
+    def test_ptp(self) -> None:
+        """Test the ptp atom.
+        """
+        a = np.array([[10., -10., 3.0], [6., 0., -1.5]])
+        expr = cp.ptp(a)
+        assert expr.is_nonneg()
+        assert expr.shape == ()
+        assert np.isclose(expr.value, 20.)
+
+        expr = cp.ptp(a, axis=0)
+        assert expr.is_nonneg()
+        assert expr.shape == (3,)
+        assert np.allclose(expr.value, np.array([4, 10, 4.5]))
+
+        expr = cp.ptp(a, axis=1)
+        assert expr.is_nonneg()
+        expr.shape == (2,)
+        assert np.allclose(expr.value, np.array([20., 7.5]))
+
+        expr = cp.ptp(a, 0, True)
+        assert expr.is_nonneg()
+        assert expr.shape == (1, 3)
+        assert np.allclose(expr.value, np.array([[4, 10, 4.5]]))
+
+        expr = cp.ptp(a, 1, True)
+        assert expr.is_nonneg()
+        assert expr.shape == (2, 1)
+        assert np.allclose(expr.value, np.array([[20.], [7.5]]))
+
+        x = cp.Variable(10)
+        expr = cp.ptp(x)
+        assert expr.curvature == 'CONVEX'
+
+    def test_stats(self) -> None:
+        """Test the mean, std, var atoms.
+        """
+        a = np.array([[10., 10., 3.0], [6., 0., 1.5]])
+        expr_mean = cp.mean(a)
+        expr_var = cp.var(a)
+        expr_std = cp.std(a)
+        assert expr_mean.is_nonneg()
+        assert expr_var.is_nonneg()
+        assert expr_std.is_nonneg()
+
+        assert np.isclose(a.mean(), expr_mean.value)
+        assert np.isclose(a.var(), expr_var.value)
+        assert np.isclose(a.std(), expr_std.value)
+
+        for ddof in [0, 1]:
+            expr_var = cp.var(a, ddof=ddof)
+            expr_std = cp.std(a, ddof=ddof)
+
+            assert np.isclose(a.var(ddof=ddof), expr_var.value)
+            assert np.isclose(a.std(ddof=ddof), expr_std.value)
+
+        for axis in [0, 1]:
+            for keepdims in [True, False]:
+                expr_mean = cp.mean(a, axis=axis, keepdims=keepdims)
+                # expr_var = cp.var(a, axis=axis, keepdims=keepdims)
+                expr_std = cp.std(a, axis=axis, keepdims=keepdims)
+
+                assert expr_mean.shape == a.mean(axis=axis, keepdims=keepdims).shape
+                # assert expr_var.shape == a.var(axis=axis, keepdims=keepdims).shape
+                assert expr_std.shape == a.std(axis=axis, keepdims=keepdims).shape
+
+                assert np.allclose(a.mean(axis=axis, keepdims=keepdims), expr_mean.value)
+                # assert np.allclose(a.var(axis=axis, keepdims=keepdims), expr_var.value)
+                assert np.allclose(a.std(axis=axis, keepdims=keepdims), expr_std.value)
+
     def test_partial_optimize_dcp(self) -> None:
         """Test DCP properties of partial optimize.
         """
@@ -1112,18 +1310,15 @@ class TestAtoms(BaseTest):
         """Test that norm1 and normInf match definition for matrices.
         """
         A = np.array([[1, 2], [3, 4]])
-        print(A)
         X = Variable((2, 2))
         obj = Minimize(cp.norm(X, 1))
         prob = cp.Problem(obj, [X == A])
         result = prob.solve(solver=cp.SCS)
-        print(result)
         self.assertAlmostEqual(result, cp.norm(A, 1).value, places=3)
 
         obj = Minimize(cp.norm(X, np.inf))
         prob = cp.Problem(obj, [X == A])
         result = prob.solve(solver=cp.SCS)
-        print(result)
         self.assertAlmostEqual(result, cp.norm(A, np.inf).value, places=3)
 
     def test_indicator(self) -> None:
@@ -1191,13 +1386,17 @@ class TestAtoms(BaseTest):
             -result, 4 * np.log(scipy.stats.norm.cdf(2)), places=None, delta=1e-2
         )
 
-    def test_scalar_product(self) -> None:
+    def test_vdot(self) -> None:
         """Test scalar product.
         """
         p = np.ones((4,))
         v = cp.Variable((4,))
 
-        p = np.ones((4,))
+        obj = cp.Minimize(cp.vdot(v, p))
+        prob = cp.Problem(obj, [v >= 1])
+        prob.solve(solver=cp.SCS)
+        assert np.allclose(v.value, p)
+
         obj = cp.Minimize(cp.scalar_product(v, p))
         prob = cp.Problem(obj, [v >= 1])
         prob.solve(solver=cp.SCS)
@@ -1208,11 +1407,51 @@ class TestAtoms(BaseTest):
         v = cp.Variable((4,))
 
         p.value = np.ones((4,))
+        obj = cp.Minimize(cp.vdot(v, p))
+        prob = cp.Problem(obj, [v >= 1])
+        prob.solve(solver=cp.SCS)
+        assert np.allclose(v.value, p.value)
+
         obj = cp.Minimize(cp.scalar_product(v, p))
         prob = cp.Problem(obj, [v >= 1])
         prob.solve(solver=cp.SCS)
         assert np.allclose(v.value, p.value)
 
+    def test_outer(self) -> None:
+        """Test the outer atom.
+        """
+        a = np.ones((3,))
+        b = Variable((2,))
+        expr = cp.outer(a, b)
+        self.assertEqual(expr.shape, (3, 2))
+
+        # Test with parameter
+        c = Parameter((2,))
+        expr = cp.outer(c, a)
+        self.assertEqual(expr.shape, (2, 3))
+
+        d = np.ones((4,))
+        expr = cp.outer(a, d)
+        true_val = np.outer(a, d)
+        assert np.allclose(expr.value, true_val, atol=1e-1)
+
+        # Test with scalars
+        assert np.allclose(np.outer(3, 2), cp.outer(3, 2).value)
+        assert np.allclose(np.outer(3, d), cp.outer(3, d).value)
+
+        # Test with matrices
+        A = np.arange(4).reshape((2, 2))
+        np.arange(4, 8).reshape((2, 2))
+
+        with pytest.raises(ValueError, match="x must be a vector"):
+            cp.outer(A, d)
+        with pytest.raises(ValueError, match="y must be a vector"):
+            cp.outer(d, A)
+
+        # allow 2D inputs once row-major flattening is the default
+        assert np.allclose(cp.vec(np.array([[1, 2], [3, 4]])).value, np.array([1, 3, 2, 4]))
+
+
     def test_conj(self) -> None:
         """Test conj.
         """
